/*
 * Ben NanoNote LCD initialisation, based on uboot-xburst and xburst-tools.
 *
 * Copyright (C) 2001-2002 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 * Copyright (C) 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "board-display.h"

#include "jzlcd.h"
#include "sdram.h"
#include "board.h"

extern vidinfo_t panel_info;
static uint32_t lcd_base;

static uint16_t get_line_length()
{
	return ALIGN((panel_info.jz_fb->w * panel_info.jz_fb->bpp) / 8, sizeof(uint32_t));
}

static uint32_t *get_pixel32(uint16_t h, uint16_t v)
{
	return (uint32_t *) (lcd_base + v * get_line_length()) + h;
}

static uint16_t *get_pixel16(uint16_t h, uint16_t v)
{
	return (uint16_t *) (lcd_base + v * get_line_length()) + h;
}

static uint8_t *get_pixel8(uint16_t h, uint16_t v)
{
	return (uint8_t *) (lcd_base + v * get_line_length()) + h;
}

static uint8_t *get_pixel4(uint16_t h, uint16_t v)
{
	return (uint8_t *) (lcd_base + v * get_line_length()) + h / 2;
}

static inline uint16_t div(uint16_t num, uint16_t denom, uint16_t scale)
{
	return (scale * num) / denom;
}

static uint32_t pixel(uint8_t r, uint8_t g, uint8_t b, uint8_t rmax, uint8_t gmax, uint8_t bmax, uint8_t rshift, uint8_t gshift, uint8_t bshift)
{
	return (div(r, 255, rmax) << rshift) | (div(g, 255, gmax) << gshift) | (div(b, 255, bmax) << bshift);
}

uint32_t get_bitmap_value(uint16_t x, uint32_t value)
{
	return (value >> (((panel_info.jz_fb->w - 1 - x) * 32) / panel_info.jz_fb->w)) % 2 ? 0xffffffff : 0;
}

static void get_colour(uint16_t h, uint16_t v, uint8_t rgb[], uint16_t pixel_type)
{
	uint16_t v_max = panel_info.jz_fb->h - 1;
	uint16_t h_max = panel_info.jz_fb->w - 1;

	rgb[(pixel_type - 1) % 3] = div(h, h_max, 255);
	rgb[pixel_type % 3] = div(v, v_max, 255);
	rgb[(pixel_type + 1) % 3] = (rgb[0] + rgb[1]) / 2;
}

static void set_pixel32(uint16_t h, uint16_t v, uint32_t value)
{
	uint32_t *pix = get_pixel32(h, v);
	*pix = value;
}

static void set_pixel16_565(uint16_t h, uint16_t v, uint32_t value)
{
	uint16_t *pix = get_pixel16(h, v);
	*pix = (uint16_t) value;
}

static void set_pixel8(uint16_t h, uint16_t v, uint32_t value)
{
	uint8_t *pix = get_pixel8(h, v);
	*pix = (uint8_t) value;
}

static void set_pixel4(uint16_t h, uint16_t v, uint32_t value)
{
	uint8_t *pix = get_pixel4(h, v);
	uint8_t mask = h & 1 ? 0xf0 : 0x0f;
	*pix = (*pix & mask) | ((uint8_t) value);
}

void set_pixel(uint16_t h, uint16_t v, uint32_t value)
{
	switch (panel_info.jz_fb->bpp)
	{
		case 32:
		set_pixel32(h, v, value);
		break;

		case 8:
		set_pixel8(h, v, value);
		break;

		case 4:
		set_pixel4(h, v, value);
		break;

		case 16:
		default:
		set_pixel16_565(h, v, value);
		break;
	}
}
static void test_pixel32(uint16_t h, uint16_t v, uint16_t pixel_type)
{
	uint8_t rgb[3];

	get_colour(h, v, rgb, pixel_type);
	set_pixel32(h, v, pixel(rgb[0], rgb[1], rgb[2], 255, 255, 255, 16, 8, 0) | get_bitmap_value(h, pixel_type));
}

static void test_pixel16_565(uint16_t h, uint16_t v, uint16_t pixel_type)
{
	uint8_t rgb[3];

	get_colour(h, v, rgb, pixel_type);
	set_pixel16_565(h, v, pixel(rgb[0], rgb[1], rgb[2], 31, 63, 31, 11, 5, 0) | get_bitmap_value(h, pixel_type));
}

static void test_pixel8(uint16_t h, uint16_t v, uint16_t pixel_type)
{
	uint8_t rgb[3];

	get_colour(h, v, rgb, pixel_type);
	set_pixel8(h, v, pixel(rgb[0], rgb[1], rgb[2], 7, 7, 3, 5, 2, 0) | get_bitmap_value(h, pixel_type));
}

static void test_pixel4(uint16_t h, uint16_t v, uint16_t pixel_type)
{
	uint8_t rgb[3];

	get_colour(h, v, rgb, pixel_type);
	set_pixel4(h, v, pixel(rgb[0], rgb[1], rgb[2], 1, 2, 1, 3, 1, 0) | get_bitmap_value(h, pixel_type));
}

void test_pixel(uint16_t h, uint16_t v, uint16_t pixel_type)
{
	switch (panel_info.jz_fb->bpp)
	{
		case 32:
		test_pixel32(h, v, pixel_type);
		break;

		case 8:
		test_pixel8(h, v, pixel_type);
		break;

		case 4:
		test_pixel4(h, v, pixel_type);
		break;

		case 16:
		default:
		test_pixel16_565(h, v, pixel_type);
		break;
	}
}

void clear_pixel32(uint16_t h, uint16_t v)
{
	uint32_t *pix = get_pixel32(h, v);
	*pix = 0;
}

void clear_pixel16(uint16_t h, uint16_t v)
{
	uint16_t *pix = get_pixel16(h, v);
	*pix = 0;
}

void clear_pixel8(uint16_t h, uint16_t v)
{
	uint8_t *pix = get_pixel8(h, v);
	*pix = 0;
}

void clear_pixel4(uint16_t h, uint16_t v)
{
	uint8_t *pix = get_pixel4(h, v);
	uint8_t mask = h & 1 ? 0xf0 : 0x0f;
	*pix = *pix & mask;
}

void clear_pixel(uint16_t h, uint16_t v)
{
	switch (panel_info.jz_fb->bpp)
	{
		case 32:
		clear_pixel32(h, v);
		break;

		case 8:
		clear_pixel8(h, v);
		break;

		case 4:
		clear_pixel4(h, v);
		break;

		case 16:
		default:
		clear_pixel16(h, v);
		break;
	}
}

void test_pattern()
{
	uint16_t v_max  = panel_info.jz_fb->h;
	uint16_t h_max  = panel_info.jz_fb->w;
	uint16_t v, h;

	for (v = 0; v < v_max; v += 1) {
		for (h = 0; h < h_max; h += 1) {
			test_pixel(h, v, 1);
		}
	}
}

void lcd_clear(uint32_t lcd_base)
{
	uint16_t v_max  = panel_info.jz_fb->h;
	uint16_t h_max  = panel_info.jz_fb->w;
	uint16_t v, h;
	uint32_t *pix = (uint32_t *)lcd_base;

	for (v = 0; v < v_max; v += 1) {
		for (h = 0; h < h_max; h += 1) {
			*pix++ = 0;
		}
	}
}

/* LCD initialisation. */

void lcd_init()
{
	__lcd_display_pin_init();
	__lcd_display_on();

	/* Initialise the member here since the address is otherwise invalid. */

	panel_info.jz_fb = (struct jzfb_info *) ((uint32_t) panel_info.jz_fb & ~0x80000000);

	lcd_base = lcd_ctrl_init();
	lcd_clear(lcd_base);
	lcd_enable();
}
